// Copyright Maciej Sobczak 2008-2022.
// This file is part of YAMI4.
// See the package-level LICENSE.txt file.

#ifndef YAMICORE_ALLOCATOR_H_INCLUDED
#define YAMICORE_ALLOCATOR_H_INCLUDED

#include "dll.h"

#include <cstddef>

namespace yami
{
    
namespace core
{

/// \brief Common interface for the custom memory allocator.
class DLL allocator
{
public:
    virtual void * allocate(std::size_t requested_size) = 0;

    virtual void deallocate(const void * p) = 0;
};

/// \brief Standard (malloc/free), default allocator.
class DLL standard_allocator : public allocator
{
public:
    virtual void * allocate(std::size_t requested_size);

    virtual void deallocate(const void * p);
};

/// \brief Non-locking (thread-unsafe) block-based allocator.
///
/// The block-based allocator operates within the given block of memory.
/// This allocator uses a straightforward implementation with lists of free blocks,
/// which are coalesced at deallocation time.
/// It is appropriate for repeated alloc/dealloc patterns,
/// but is vulnerable to fragmentation with irregular request sequences.
/// This allocator should be protected by a locking wrapper, if intended for use
/// in a multithreading environment.
class DLL non_locked_allocator : public allocator
{
public:
    non_locked_allocator();

    void set_working_area(void * buf, std::size_t size);

    virtual void * allocate(std::size_t requested_size);

    virtual void deallocate(const void * p);

    void get_free_size(std::size_t & biggest, std::size_t & all) const;

private:
    non_locked_allocator(const non_locked_allocator &);
    void operator=(const non_locked_allocator &);

    void * base_;
    std::size_t size_;

    void * first_free_segment_;
};

} // namespace core

} // namespace yami

#endif // YAMICORE_ALLOCATOR_H_INCLUDED
